from datetime import datetime, timedelta

import databases


async def get_latest_stats(dbconn: databases.Database):
    result = await dbconn.fetch_one(
        "SELECT * FROM aqi_log ORDER BY event_time DESC LIMIT 1"
    )
    if result:
        return result[0], result[1], result[2], result[3]
    else:
        return 0, 0, 0, 0


async def get_all_stats(dbconn: databases.Database):
    result = await dbconn.fetch_all("SELECT * FROM aqi_log ORDER BY event_time ASC")
    return result


async def clean_old(dbconn: databases.Database, retention_minutes: int):
    last_week = datetime.now() - timedelta(minutes=retention_minutes)
    last_week_timestamp = int(last_week.timestamp())
    await dbconn.execute(
        "DELETE FROM aqi_log WHERE event_time < :last_week_timestamp",
        values={"last_week_timestamp": last_week_timestamp},
    )


async def add_entry(
    dbconn: databases.Database, event_time, epa_aqi_pm25, raw_pm25, raw_pm10
):
    formatted_time = int(event_time.timestamp())
    await dbconn.execute(
        query="INSERT INTO aqi_log VALUES (:formatted_time, :epa_aqi_pm25, :raw_pm25, :raw_pm10)",
        values={
            "formatted_time": formatted_time,
            "epa_aqi_pm25": epa_aqi_pm25,
            "raw_pm25": raw_pm25,
            "raw_pm10": raw_pm10,
        },
    )


async def create_tables(dbconn: databases.Database):
    # Create table
    await dbconn.execute(
        """CREATE TABLE IF NOT EXISTS aqi_log (event_time integer, epa_aqi_pm25 real, raw_pm25 real, raw_pm10 real)"""
    )
    await dbconn.execute(
        """CREATE INDEX IF NOT EXISTS aqi_eventtime ON aqi_log (event_time)"""
    )



import os.path
from typing import Optional
from dataclasses import dataclass
from serde import serde
from serde.toml import to_toml, from_toml

DEFAULT_CONFIG_PATH = "~/.aqimon/config"
DEFAULT_DB_PATH = "~/.aqimon/db.sqlite"


@serde
@dataclass(frozen=True)
class Config:
    database_path: str
    poll_frequency_sec: int
    retention_minutes: int

    reader_type: str

    # Nova PM properties
    usb_path: str
    usb_sleep_time_sec: int
    sample_count_per_read: int


DEFAULT_CONFIG = Config(
    database_path=os.path.expanduser(DEFAULT_DB_PATH),
    poll_frequency_sec=60 * 15,  # Every 15 minutes
    retention_minutes=60 * 24 * 7,  # 1 week
    reader_type="NOVAPM",
    usb_path="/dev/ttyUSB0",
    usb_sleep_time_sec=5,
    sample_count_per_read=5,
)


def _load_config(path: str) -> Config:
    with open(path, "r") as file:
        return from_toml(Config, file.read())


def save_config(config: Config, path: str):
    with open(path, "w") as file:
        file.write(to_toml(config))


def get_config(passed_config_path: Optional[str]) -> Config:
    if passed_config_path and os.path.exists(passed_config_path):
        return _load_config(passed_config_path)
    elif not passed_config_path and os.path.exists(DEFAULT_CONFIG_PATH):
        return _load_config(DEFAULT_CONFIG_PATH)
    else:
        return DEFAULT_CONFIG



import json
from fastapi import FastAPI, Request
from fastapi.templating import Jinja2Templates
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fastapi_utils.tasks import repeat_every
import uvicorn
import databases
from pathlib import Path
from datetime import datetime
from .database import (
    get_latest_stats,
    get_all_stats,
    add_entry,
    create_tables,
    clean_old,
)
from .read import AqiRead, Reader, ReaderState
from .read.mock import MockReader
from .read.novapm import NovaPmReader
from . import aqi_common
from .config import Config, get_config

app = FastAPI()
config = get_config(None)
app.mount("/static", StaticFiles(directory="aqimon/static"), name="static")

Path(config.database_path).parent.mkdir(parents=True, exist_ok=True)
database = databases.Database(f"sqlite+aiosqlite:///{config.database_path}")
templates = Jinja2Templates(directory="aqimon/templates")


def _get_reader(conf: Config) -> Reader:
    if conf.reader_type == "MOCK":
        return MockReader()
    elif conf.reader_type == "NOVAPM":
        return NovaPmReader(
            usb_path=conf.usb_path,
            iterations=conf.sample_count_per_read,
            sleep_time=conf.usb_sleep_time_sec,
        )
    else:
        raise Exception("Invalid reader type specified")


reader = _get_reader(config)
reader_state = ReaderState(False, None)


@app.on_event("startup")
async def database_connect():
    await database.connect()
    await create_tables(database)


@app.on_event("startup")
@repeat_every(seconds=5)
async def read_from_device() -> None:
    global reader_state
    try:
        result: AqiRead = reader.read()
        event_time = datetime.now()
        epa_aqi_pm25 = aqi_common.calculate_epa_aqi(result.pmtwofive)
        await add_entry(
            dbconn=database,
            event_time=event_time,
            epa_aqi_pm25=epa_aqi_pm25,
            raw_pm25=result.pmtwofive,
            raw_pm10=result.pmten,
        )
        await clean_old(dbconn=database, retention_minutes=config.retention_minutes)
        reader_state = ReaderState(True, None)
    except Exception as e:
        reader_state = ReaderState(False, e)


@app.on_event("shutdown")
async def database_disconnect():
    await database.disconnect()


def convert_all_to_view_dict(results):
    view = {}
    view["t"] = []
    view["epa"] = []
    view["pm25"] = []
    view["pm10"] = []
    for x in results:
        view["t"].append(datetime.fromtimestamp(int(x[0])).strftime("%m-%d %H:%M"))
        view["epa"].append(x[1])
        view["pm25"].append(x[2])
        view["pm10"].append(x[3])
    return view


@app.post("/add")
async def add_new_entry() -> AqiRead:
    data = reader.read()
    epa_aqi_pm25 = aqi_common.calculate_epa_aqi(data.pmtwofive)
    await add_entry(
        dbconn=database,
        event_time=datetime.now(),
        epa_aqi_pm25=epa_aqi_pm25,
        raw_pm25=data.pmtwofive,
        raw_pm10=data.pmten,
    )
    return data


@app.get("/", response_class=HTMLResponse)
async def home(request: Request):
    return templates.TemplateResponse(
        "index.html",
        {
            "request": request,
        },
    )


@app.get("/api/alldata")
async def all_data():
    all_stats = await get_all_stats(database)
    all_json = convert_all_to_view_dict(all_stats)
    return all_json


@app.get("/api/status")
async def status():
    return {
        "reader_alive": reader_state.alive,
        "reader_exception": str(reader_state.last_exception)
        if reader_state is not None
        else None,
    }


def start():
    uvicorn.run(app, host="0.0.0.0", port=8000)



import random
from aqimon.read import AqiRead


class MockReader:
    def read(self) -> AqiRead:
        pm25: float = round(random.uniform(0.0, 500.4), 2)
        pm10: float = round(random.uniform(0.0, 300.0), 2)
        return AqiRead(pmtwofive=pm25, pmten=pm10)



from typing import Protocol, Optional
from dataclasses import dataclass


@dataclass(frozen=True)
class AqiRead:
    pmtwofive: float
    pmten: float


class Reader(Protocol):
    def read(self) -> AqiRead:
        pass


@dataclass(frozen=True)
class ReaderState:
    alive: bool
    last_exception: Optional[Exception]



from typing import List, Tuple

AQI: List[Tuple[int, int]] = [
    (0, 50),
    (51, 100),
    (101, 150),
    (151, 200),
    (201, 300),
    (301, 400),
    (401, 500),
]
PM_25: List[Tuple[float, float]] = [
    (0.0, 12.0),
    (12.1, 35.4),
    (35.5, 55.4),
    (55.5, 150.4),
    (150.5, 250.4),
    (250.5, 350.4),
    (350.5, 500.4),
]
COLORS = ["green", "yellow", "orange", "red", "purple", "dark-red", "dark-red"]


def get_level_from_pm25(pm25: float) -> int:
    for i, pair in enumerate(PM_25):
        if pair[0] <= pm25 <= pair[1]:
            return i
    raise ValueError("Invalid PM value")


def get_color_from_level(level: int) -> str:
    return COLORS[level]


def calculate_epa_aqi(pm_25_read: float) -> int:
    for i, pm_range in enumerate(PM_25):
        if pm_range[0] <= pm_25_read <= pm_range[1]:
            aqi_low = AQI[i][0]
            aqi_high = AQI[i][1]
            pm_low = pm_range[0]
            pm_high = pm_range[1]
            epa = ((aqi_high - aqi_low) / (pm_high - pm_low)) * (
                pm_25_read - pm_low
            ) + aqi_low
            return round(epa)
    return -1



from aqimon import usb
from . import AqiRead
import serial
import time
from statistics import mean


class NovaPmReader:
    def __init__(self, usb_path: str, iterations: int = 5, sleep_time: int = 60):
        self.usb_path = usb_path
        self.iterations = iterations
        self.sleep_time = sleep_time

    def read(self) -> AqiRead:
        return self._power_saving_read()

    def _power_saving_read(self) -> AqiRead:
        try:
            usb.turn_on_usb()
            time.sleep(5)
        except usb.UhubCtlNotInstalled:
            pass
        result = self._averaged_read()
        try:
            usb.turn_off_usb()
            time.sleep(5)
        except usb.UhubCtlNotInstalled:
            pass

        return AqiRead(result.pmtwofive, result.pmten)

    def _averaged_read(self) -> AqiRead:
        pm25_reads = []
        pm10_reads = []

        for x in range(self.iterations):
            data = self._read()
            pm25_reads.append(data.pmtwofive)
            pm10_reads.append(data.pmten)
            time.sleep(self.sleep_time)

        avg_pm25 = mean(pm25_reads)
        avg_pm10 = mean(pm10_reads)

        return AqiRead(pmtwofive=avg_pm25, pmten=avg_pm10)

    def _read(self) -> AqiRead:
        ser = serial.Serial(self.usb_path)
        data = ser.read(10)
        pmtwofive = int.from_bytes(data[2:4], byteorder="little") / 10
        pmten = int.from_bytes(data[4:6], byteorder="little") / 10
        return AqiRead(pmtwofive, pmten)



import subprocess
import pathlib

UHUB_CTL_PATH = "uhubctl"


class UhubCtlNotInstalled(Exception):
    def __init__(self):
        super().__init__("Unable to find uhubctl on path")


def turn_on_usb():
    if not _uhubctl_installed():
        raise UhubCtlNotInstalled()
    subprocess.run([UHUB_CTL_PATH, "-l", "1-1", "-a", "on", "-p", "2"], check=True)


def turn_off_usb():
    if not _uhubctl_installed():
        raise UhubCtlNotInstalled()
    subprocess.run([UHUB_CTL_PATH, "-l", "1-1", "-a", "off", "-p", "2"], check=True)


def _uhubctl_installed() -> bool:
    return pathlib.Path(UHUB_CTL_PATH).exists()
